"""
Letter to int mapping dictionaries. Characters (a-z) are mapped to 1 - 26.

ordered_lett_mapping_dict:
Mapping of letters according to their English ordering, e.g. a -> 1, b -> 2, ..., z -> 26

english_lett_freq_mapping_dict:
Mapping of letters according to their English frequency, with more frequent letters mapped to the middle of
the range 1 - 26, and less frequent words mapped to the edges of the range. Since e, t, and a are the most
frequent letters, they are mapped to 14, 13, 1nd 15 respectively. Since q, j, and z are the least frequent,
they are mapped to 2, 26, and 1 respectively.

english_lett_freq_mapping_dict_aesthetic:
Variant on the mapping described in english_lett_freq_mapping_dict where I randomly swapped adjacent pairs
and triples of the letter mappings and accepted a new mapping if it lowered the standard deviation of the
letter to number mapping of my example corpus (2500 word sample from English Wikipedia articles). The
mapping below is the result of 1e6 iterations of adjacent pair swaps and 1e5 adjacent triple swaps. Adjacent
triples were randomly permuted when they were swapped. Note that this version doesn't optimize the
standard deviation across corpuses, the plain frequency version performs better there. I kept this mapping
for aesthetic value only.

"""
ordered_lett_mapping_dict = {
    "a": 1,
    "b": 2,
    "c": 3,
    "d": 4,
    "e": 5,
    "f": 6,
    "g": 7,
    "h": 8,
    "i": 9,
    "j": 10,
    "k": 11,
    "l": 12,
    "m": 13,
    "n": 14,
    "o": 15,
    "p": 16,
    "q": 17,
    "r": 18,
    "s": 19,
    "t": 20,
    "u": 21,
    "v": 22,
    "w": 23,
    "x": 24,
    "y": 25,
    "z": 26,
}

english_lett_freq_mapping_dict = {
    "e": 14,
    "t": 13,
    "a": 15,
    "o": 12,
    "i": 16,
    "n": 11,
    "s": 17,
    "r": 10,
    "h": 18,
    "d": 9,
    "l": 19,
    "u": 8,
    "c": 20,
    "m": 7,
    "f": 21,
    "y": 6,
    "w": 22,
    "g": 5,
    "p": 23,
    "b": 4,
    "v": 24,
    "k": 3,
    "x": 25,
    "q": 2,
    "j": 26,
    "z": 1,
}

english_lett_freq_mapping_dict_aesthetic = {
    "e": 15,
    "t": 13,
    "a": 14,
    "o": 17,
    "i": 16,
    "n": 12,
    "s": 11,
    "r": 18,
    "h": 19,
    "d": 9,
    "l": 10,
    "u": 8,
    "c": 20,
    "m": 21,
    "f": 7,
    "y": 5,
    "w": 6,
    "g": 23,
    "p": 22,
    "b": 24,
    "v": 25,
    "k": 4,
    "x": 3,
    "q": 1,
    "j": 2,
    "z": 26,
}

mapping_dicts = {
    "ordered": ordered_lett_mapping_dict,
    "frequency": english_lett_freq_mapping_dict,
    "aesthetic": english_lett_freq_mapping_dict_aesthetic,
}


inverse_mapping_dicts = {}
for key, mapping_dict in mapping_dicts.items():
    inverse_mapping_dicts[key] = {v: k for k, v in mapping_dict.items()}



import pathlib
from setuptools import setup
from grampyx.__init__  import __version__


PARENT_DIR = pathlib.Path(__file__).parent
README = (PARENT_DIR / "README.md").read_text()

setup(
    name="grampyx",
    version=__version__,
    description="Convert text to image",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/cbattle12/grampyx",
    author="Christopher Battle",
    author_email="christopher.g.battle@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
    ],
    python_requires=">=3.5",
    packages=["grampyx"],
    include_package_data=True,
    install_requires=["numpy>=1.15.0"],
)



import numpy as np
import warnings
import re
from typing import Optional

from grampyx.error_handling import NoAlphabetCharactersWarning, PixelValuesWarning
from grampyx._base_functions import ARRAY_DIM, _word2pic, _pic2word


def grams2pix(
    words: str,
    mapping: str = "aesthetic",
    pictype: str = "gradient",
    compress: bool = False,
    separator: Optional[str] = None,
    n: Optional[int] = None,
) -> np.ndarray:
    """
    Convert string of words to image. Each word is converted to an ARRAY-DIM x ARRAY_DIM image which is embedded in
    a larger n x n image array, ordered from left to right, top to bottom

    :param words: string of words separated by separator (defaults to whitespace)
    :param mapping: char to pixel mapping dictionary. Valid values are ordered, frequency, aesthetic
    :param pictype: type of picture. Valid values are punchcard (binary) or gradient (interpolated grayscale)
    :param compress: compress string bool. If True, shorten word by removing letters per their ordering in the
            mapping dict for words > 28 characters. If False, map only the first 28 characters of the word.
    :param separator: word separator, used as arg for str.split()
    :param n: dimension of square image to return (n x n). If the number of words < n x n, the extra space is
                zero-padded. Default behavior is to take max n where n x n < number of words.
    :return: mapped image
    """
    if not re.search("[a-zA-Z]", words):
        warnings.warn(
            f"No English alphabet characters found in input string, returning {ARRAY_DIM} x {ARRAY_DIM} "
            f"zero-array",
            NoAlphabetCharactersWarning,
        )
        return np.zeros((ARRAY_DIM, ARRAY_DIM))

    word_list = words.split(separator)

    # Use smallest square that contains all words.
    if not n:
        n = int(np.sqrt(len(word_list))) + 1

    pic = np.zeros((ARRAY_DIM * n, ARRAY_DIM * n))
    # Fill array left to right, top to bottom # j/n, cut off word list first, column row names, add black
    for word_idx, word in enumerate(word_list[: n ** 2]):
        column = word_idx % n
        row = int(word_idx / n)
        pic[
            row * ARRAY_DIM : (row + 1) * ARRAY_DIM,
            column * ARRAY_DIM : (column + 1) * ARRAY_DIM,
        ] = _word2pic(word, mapping=mapping, pictype=pictype, compress=compress)
    return pic


def pix2grams(pic: np.ndarray, mapping: str = "aesthetic", separator: str = " ") -> str:
    """
    Convert picture to string of words. Scans over the image from left to right, top to bottom, and converts
    ARRAY_DIM x ARRAX_DIM blocks to words

    :param pics: image to convert. Images where all pixels < 1 will return all 0s. Image must be square
    :param mapping: char to pixel mapping dictionary. Valid values are ordered, frequency, aesthetic
    :param separator: word separator
    :return: mapped string
    """

    if (pic < 1).all():
        warnings.warn(
            f"All image pixel values < 1, returning empty string. Please rescale your image to contain values"
            " >= 1 to generate text.",
            PixelValuesWarning,
        )
        return ""

    words = []
    length = int(pic.shape[0] / ARRAY_DIM)
    width = int(pic.shape[1] / ARRAY_DIM)
    # Iterate over array left to right, top to bottom
    for sub_arr_idx in range(width * length):
        column = sub_arr_idx % width
        row = int(sub_arr_idx / length)
        words.append(
            _pic2word(
                pic[
                    row * ARRAY_DIM : (row + 1) * ARRAY_DIM,
                    column * ARRAY_DIM : (column + 1) * ARRAY_DIM,
                ],
                mapping=mapping,
            )
        )
    return separator.join(words).strip(separator)



import numpy as np

from grampyx.mappings import mapping_dicts, inverse_mapping_dicts

ARRAY_DIM = 28
VALID_PICTURE_TYPES = ["gradient", "punchcard"]
# Contains pre-computed gradients for gradient images (it's faster this way)
GRADIENT_DICT = {
    i: np.linspace(1 / ARRAY_DIM, (ARRAY_DIM - 1) / ARRAY_DIM, i)
    for i in range(1, ARRAY_DIM)
}


def _word2pic(
    word: str,
    mapping: str = "aesthetic",
    pictype: str = "gradient",
    compress: bool = False,
) -> np.ndarray:
    """
    Convert string to ARRAY_DIM x ARRAY_DIM image

    :param word: string to convert, only letters a-z are mapped
    :param mapping: char to pixel mapping dictionary. Valid values are ordered, frequency, aesthetic
    :param pictype: type of picture. Valid values are punchcard (binary) or gradient (interpolated grayscale)
    :param compress: compress string bool. If True, shorten word by removing letters per their ordering in the
            mapping dict for words > 28 characters. If False, map only the first 28 characters of the word.
    :return: mapped image
    """
    if pictype not in VALID_PICTURE_TYPES:
        raise ValueError(f"Invalid pictype - valid pictypes are: {VALID_PICTURE_TYPES}")
    if mapping in mapping_dicts.keys():
        mapping_dict = mapping_dicts[mapping]
    else:
        raise KeyError(f"Invalid mapping - valid mappings are: {mapping_dicts.keys()}")

    word = word.lower().strip()
    if len(word) > ARRAY_DIM:
        if compress:
            for letter in mapping_dict.keys():
                word = word.replace(letter, "")
                if len(word) <= ARRAY_DIM:
                    break
            # If the word length is still > ARRAY_DIM, e.g. due to special characters, no mapping is possible
            if len(word) > ARRAY_DIM or len(word) == 0:
                return np.zeros((ARRAY_DIM, ARRAY_DIM))
        else:
            word = word[:ARRAY_DIM]

    pic = np.zeros((ARRAY_DIM, ARRAY_DIM))
    start_idx = int((ARRAY_DIM - len(word)) / 2)  # Centering
    previous_y = np.nan
    for x, letter in enumerate(word):
        y = mapping_dict.get(letter, np.nan)
        if np.isnan(y):
            previous_y = y
            continue
        pic[y, start_idx + x] = 1
        if pictype == "gradient":
            # Add grayscale gradients between consecutive pixels in the binary image. Gradient
            # goes from dark to light, from previous pixel y-value to next pixel y-value
            if not np.isnan(previous_y):
                diff = y - previous_y
                if diff > 0:
                    pic[previous_y:y, start_idx + x] = GRADIENT_DICT[int(diff)]
                if diff < 0:
                    pic[previous_y:y:-1, start_idx + x] = GRADIENT_DICT[
                        int(previous_y - y)
                    ]
        previous_y = y
    return pic


def _pic2word(pic: np.ndarray, mapping: str = "aesthetic") -> str:
    """
    Convert image to string

    :param pic: image to convert. Images where all pixels < 1 will return all 0s. Image must be square with side <= 28
    :param mapping: char to pixel mapping dictionary. Valid values are ordered, frequency, aesthetic
    :return: mapped string
    """

    if pic.shape[0] != pic.shape[1]:
        raise ValueError(
            f"First two array dimensions are not equal. Image arrays must be square."
        )
    if pic.shape[0] > ARRAY_DIM:
        raise ValueError(
            f"First array dimension is {pic.shape[0]}, must be < {ARRAY_DIM}"
        )
    if mapping in inverse_mapping_dicts.keys():
        mapping_dict = inverse_mapping_dicts[mapping]
    else:
        raise KeyError(f"Invalid mapping - valid mappings are: {mapping_dicts.keys()}")

    pic.astype(int).clip(0, 1, out=pic)  # convert image to binary
    letter_list = []
    # Note only the first pixel with a value of 1 is mapped to a letter
    for i in range(pic.shape[0]):
        column = pic[:, i]
        y_idx_column = column.nonzero()[0]
        if y_idx_column.size:
            letter = mapping_dict.get(y_idx_column[0], "")
            letter_list.append(letter)
    return "".join(letter_list)
